class TopLevelParse:
	def __init__(self):
		self.parser_stacks = []
		self.parser_objs = {}
		self.input_stream_indices = IndexTree() # TODO: turn this into a tree
		self.input_index_tree_path = [0]
		self.current_parse_depth = 0 # Used by the AST handling code.
		self.starting_input_index = []
		self.input_token_map = {}
		self.input_ptr = None
		# Technically, 1 virtual parse function + 1 h_do_parse + 1 perform_lowlevel_parse (if packrat)
		# TODO: len(self.parser_stacks[-1]) and self.current_parse_depth serve the same purpose. avoid multiple sources of truth
		# TODO: stack of trees to push/pop when switching parser stacks

		self.unclaimed_mem_use = 0
		# Holds 32 characters starting at state->input_stream[index], used by the GUI
		self.current_input_chunk = ''
		self.current_parser_env = ''
		# We save/push in perform_lowlevel_parse, but this is used to display them ahead of time
		self.h_do_parse_parser = None
		self.vt_types = None
		self.parser_decombinator = None

		# Counts stack pushes/pops to determine if stopping is needed for hammer-parse-after-apply
		self.ast_stack_index = None
		# Holds a reference to the relevant parser stack
		self.ast_selected_stack = None
		# Holds parse states down the stack from when hammer-parse-apply or hammer-parse-step-to-result was executed
		self.parse_state_stack = [] # TODO: remove
		self.current_token_start = 0 # TODO: remove
		self.current_token_end = 0 #TODO: remove

	def init_parser(self):
		self.vt_types = VTTypes()
		self.parser_decombinator = ParserDecombinator(self.vt_types)

	# Called from h_packrat_parse()'s handler, where the parse state and arena get initialized
	def enter_h_packrat_parse(self, parser):
		parser_stack = ParserStack(None, None)
		self.parser_stacks.append(parser_stack)
		self.current_parse_depth = -1 # entering h_do_parse increments this by 1. To have it 0-indexed, it needs to be initialized to -1
		return 0

	# TODO: arena parameter is useless
	def enter_h_do_parse(self, parse_state, arena, parser):
		parser_stack = self.peek_parserstack()
		try:
			parser_obj = self.parser_objs[parser]
		except KeyError:
			parser_obj = Parser(None, parser)
			self.parser_objs[parser] = parser_obj
		self.h_do_parse_parser = parser_obj # TODO: current_parser_env should be set here instead too
		if parser_stack.parse_state is None and parser_stack.parse_state != parse_state:
			self.first_h_do_parse_after_packrat_parse(parse_state, arena)
		self.current_parse_depth += 1
		start_index = int(parse_state['input_stream']['index'])
		self.begin_new_token(start_index)

	def return_from_h_do_parse(self, parse_state, parser, ret_val):
		parser_stack = self.peek_parserstack()
		parser_obj = self.parser_by_address(parser) # We don't try to make a new Parser() object here, because it has to have been done in the corresponding enter_h_do_parse() call
		# The rationale for handling the "current" parser separately from stack:
		# Parsers are only pushed on the stack in perform_lowlevel_parse, which doesn't get called when h_do_parse() gets the result from cache.
		# If other backends are supported, this might change to pushing/popping the stack in h_do_parse()
		self.h_do_parse_parser = parser_obj # Restore the "current" parser, otherwise it'll show the parser h_do_parse() was last called with on the GUI and backtrace

		if ret_val:
			# We'd have to subtract 1 from len(parser_stack), but since we push/pop the stack in perform_lowlevel_parse, by this point it's 1 item shorter than the depth we're at
			#token_bounds = self.input_stream_indices[len(parser_stack.p_stack)] # TODO: ParserStack.get_stack_size() or similar
			#final_token_bounds = (token_bounds[0], int(parse_state['input_stream']['index']))
			#self.input_stream_indices[len(parser_stack.p_stack)] = final_token_bounds
			end_index = int(parse_state['input_stream']['index'])
			self.finish_token(end_index, ret_val)
		else:
			# If ret_val is NULL, the parse has failed. We can be sure the (start, end) tuple won't be needed here.
			self.discard_pending_tokens()
		self.current_parse_depth -= 1

	# Called from h_do_parse()'s handler, at which point we know the addresses of the state and arena
	def first_h_do_parse_after_packrat_parse(self, parse_state, arena):
		parser_stack = self.peek_parserstack()
		parser_stack.set_state(parse_state)

	# Popping the stack of stack of parsers
	def return_from_h_packrat_parse(self):
		old_stack = self.parser_stacks.pop()
		if old_stack.depth() > 0:
			print("Warning: parser stack not empty but parse is successful?")

		top_of_parserstack = self.peek_parserstack()
		if not top_of_parserstack:
			self.current_parse_depth = -1 # Parser stack is now empty (i.e. this was the 'topmost' h_parse() invocation
		else:
			self.current_parse_depth = len(self.peek_parserstack().p_stack)-1 # This happens when a h_parse called from a semantic action or HContinuation returns (e. g. filters in pdf)

	# Memoize the parser object for this particular address, then push it on the stack
	# Returns the parser object we just initalized (or the one already existing)
	#TODO: memoize_parser method
	def enter_perform_lowlevel_parse(self, parser_addr):
		try:
			parser_obj = self.parser_objs[parser_addr]
		except KeyError:
			# Create a parser object with no name and the address of the parser
			parser_obj = Parser(None, parser_addr)
			self.parser_objs[parser_addr] = parser_obj

		parser_stack = self.peek_parserstack()
		parser_stack.push(parser_obj)
		#TODO
		parse_state = []
		parser_stack.push_parse_state(parse_state)
		if self.parser_decombinator:
			p_env = self.parser_decombinator.decompose_parser(parser_obj, self)
			self.set_parser_env(type(p_env).__name__ + " - " + str(p_env)) # TODO: pass this as data structure to frontend
		return parser_obj

	def return_from_perform_lowlevel_parse(self):
		parser_stack = self.peek_parserstack()
		parser_obj = parser_stack.pop()
		parse_state = parser_stack.pop_parse_state()
		# debug print here

	def enter_h_arena_malloc_raw(self, alloc_size):
		parser_obj = self.peek_parser()
		parser_stack = self.peek_parserstack()
		# This is probably the slowest part of the code, or maybe the overhead adds up over many calls to h_arena_malloc_raw()
		if parser_obj is not None:
			# Caveat: parser_stack is assumed not to be None if we could get a parser_obj
			parser_obj.add_mem_use(parser_stack.parse_state, alloc_size)
		elif parser_stack is not None:
			#print("Allocation of " + str(alloc_size) + " bytes without a parser on the stack. (Happens before first call perform_lowlevel_parse to or after return from that call)")
			parser_stack.unclaimed_mem_use += alloc_size
		else:
			#print("Allocation of " + str(alloc_size) + " bytes without a parser stack. (This happens before and after parse)")
			self.unclaimed_mem_use += alloc_size

	def parse_virtual(self, parser_name):
		parser_obj = self.peek_parser()
		if parser_obj.name is None:
			parser_obj.name_parser(parser_name)
		#else:
			#print("Warning: parser already named! This is a bug. old name: %s, new name: %s" % (parser_obj.name, parser_name))

	#def return_from_parse_virtual(self, parse_state, result_valid):
	#	self.current_token_end = int(parse_state['input_stream']['index']) #TODO: maybe just have ParseVirtualRetBreakpoint extract the index and pass it into the function
	#	token_bounds = (self.current_token_start, self.current_token_end)

	def begin_new_token(self, index):
		token_bounds = (index, None)
		#print("DEBUG: begin_new_token()")
		#print("DEBUG: current parse depth:", self.current_parse_depth, "len(input_index_tree_path):", len(self.input_index_tree_path), "input_index_tree_path:", self.input_index_tree_path)
		#print("DEBUG: path (truncated):", self.input_index_tree_path[0:self.current_parse_depth])
		#print("DEBUG: path:", self.input_index_tree_path)
		#print("DEBUG: tokens:", self.input_stream_indices)
		#print("DEBUG: current_parse_depth+1:", self.current_parse_depth+1, "len(input_index_tree_path):", len(self.input_index_tree_path))

		#if self.current_parse_depth == 0:
		#	self.input_stream_indices.root.begin_token(index)
		#elif len(self.input_index_tree_path) == self.current_parse_depth: # We haven't been this deep into the parse tree yet
		#	print("DEBUG: path (truncated):", self.input_index_tree_path[0:self.current_parse_depth])
		#	print("DEBUG: path:", self.input_index_tree_path)
		#	print("DEBUG: tokens:", self.input_stream_indices)
		#	leaf = self.input_stream_indices.select_by_path(self.input_index_tree_path[0:self.current_parse_depth-1])
		#	leaf.add_child_node(token_bounds)
		#	self.input_index_tree_path.append(0)
		#elif len(self.input_index_tree_path) < self.current_parse_depth:
		#	raise RuntimeError("DEBUG: index tree path shorter than parse depth. this should never happen")
		#else:
		#	print("DEBUG: path (truncated):", self.input_index_tree_path[0:self.current_parse_depth])
		#	print("DEBUG: path:", self.input_index_tree_path)
		#	print("DEBUG: tokens:", self.input_stream_indices)
		#	node = self.input_stream_indices.select_by_path(self.input_index_tree_path[0:self.current_parse_depth])
		#	node.add_child_node(token_bounds)
		#	self.input_index_tree_path[self.current_parse_depth] += 1
		#	print("DEBUG: self.current_parse_depth:", self.current_parse_depth)
		#	print("DEBUG: path after begin_new_token():", self.input_index_tree_path)
 
		if len(self.input_index_tree_path) == 1 and self.current_parse_depth == 0:
			node = self.input_stream_indices.select_by_path([0])
			try:
				node.begin_token(index)
			except ValueError:
				self.input_stream_indices.root = None
				node = IndexTreeNode(None)
				node.begin_token(index)
				self.input_stream_indices.root = node
			#node.add_child_node(token_bounds)
			#self.input_index_tree_path.append(0)
		elif len(self.input_index_tree_path) == self.current_parse_depth+1: # We returned to the parent node, and now we're applying
		# the next parser in a seq -- increment last index in the path
			node = self.input_stream_indices.select_by_path(self.input_index_tree_path[0:self.current_parse_depth]) # Select the parent of our new node
			node.add_child_node(token_bounds)
			self.input_index_tree_path[self.current_parse_depth] += 1 # Increment the last index in the path to the current node being constructed
		elif (self.current_parse_depth+1) -  len(self.input_index_tree_path) == 1:
			node = self.input_stream_indices.select_by_path(self.input_index_tree_path[0:self.current_parse_depth])
			node.add_child_node(token_bounds)
			self.input_index_tree_path.append(0)
		else:
			#print("len(self.input_index_tree_path):", len(self.input_index_tree_path), "self.current_parse_depth:", self.current_parse_depth)
			raise RuntimeError("Unexpected node path length")

		diff = (self.current_parse_depth+1) - len(self.starting_input_index)
		newindices = [None] * diff
		self.starting_input_index.extend(newindices)
		self.starting_input_index[self.current_parse_depth] = index
		#print(self.starting_input_index)

		#print("DEBUG: self.current_parse_depth:", self.current_parse_depth)
		#print("DEBUG: path after begin_new_token():", self.input_index_tree_path) # TODO: rename input_index_tree_path to next_node_path
		#print("DEBUG: tokens after begin_new_token():", self.input_stream_indices)
		#if len(self.input_stream_indices) <= self.current_parse_depth: # No tokens yet at this depth
		#	self.input_stream_indices.append([token_bounds]) # note: input_stream_indices is a list of lists, the "inner" list being bounds for tokens in a sequence
		#else: # There are already tokens at this depth (this happens when e.g. the preceding h_do_parse is the application of a h_sequence())
		#	self.input_stream_indices[self.current_parse_depth].append(token_bounds)

	# If a parser successfully returns a token, we take the list of token bounds at the current depth, check that the bounds corresponding to
	# the most recent parse result aren't filled in yet, and make note of state->input_stream.index
	def finish_token(self, index, hparseresult_addr):
		#print("DEBUG: finish_token()")
		#print("DEBUG: current parse depth:", self.current_parse_depth)
		#print("DEBUG: finished tokens:", self.input_stream_indices)
		#print("DEBUG: path:", self.input_index_tree_path)
		#print("DEBUG: path (truncated):", self.input_index_tree_path[0:self.current_parse_depth+1])
		#token = self.input_stream_indices[self.current_parse_depth][-1]
		node = self.input_stream_indices.select_by_path(self.input_index_tree_path[0:self.current_parse_depth+1])
		token = node.indices
		if token[1] is not None:
			raise RuntimeError("Attempted to determine token bounds twice:", token, "new end:", index)
		#finished_token = (token[0], index)
		#self.input_stream_indices[self.current_parse_depth][-1] = finished_token
		node.finish_token(index)
		#self.input_index_tree_path[self.current_parse_depth] += 1
		del self.input_index_tree_path[self.current_parse_depth+1:]
		token_addr = hparseresult_addr.cast(gdb.lookup_type("HParseResult").pointer())['ast']
		if token_addr:
			self.input_token_map[int(token_addr)] = {'start': self.starting_input_index[self.current_parse_depth], 'end': index, 'hparseresult': hex(hparseresult_addr)}
		#print("DEBUG: current parse depth:", self.current_parse_depth)
		#print("DEBUG: finished tokens:", self.input_stream_indices)
		#print("DEBUG: path:", self.input_index_tree_path)
		#print([str(p) for p in self.peek_parserstack().p_stack])

	# Discard a single item in the "outer" list, which is a list of tokens at that depth in the call graph (one if the parser in the stack is not a sequence, multiple if it is)
	def discard_pending_tokens(self):
		#print("DEBUG: discard_pending_tokens()")
		#print("DEBUG: current parse depth:", self.current_parse_depth)
		#print("DEBUG: path:", self.input_index_tree_path)
		#print("DEBUG: tokens:", self.input_stream_indices)
		#tokens = self.input_stream_indices.pop()
		node = self.input_stream_indices.select_by_path(self.input_index_tree_path[0:self.current_parse_depth+1])
		if type(node) is IndexTree:
			node = node.root
		#self.input_index_tree_path.pop()
		del self.input_index_tree_path[self.current_parse_depth+1:]
		#print("DEBUG: path after discard:", self.input_index_tree_path)
		if node.indices[1] is not None:
			raise RuntimeError("DEBUG: failed parse but token still has bounds:", node.indices[1])
		node.delete_all_children()
		parent = self.input_stream_indices.select_by_path(self.input_index_tree_path[0:self.current_parse_depth])
		if type(parent) is IndexTreeNode:
			parent.delete_last_child()
		else:
			parent.root = IndexTreeNode(None)
		if self.input_index_tree_path[self.current_parse_depth] > 0:
			self.input_index_tree_path[self.current_parse_depth] -= 1
		else:
			del self.input_index_tree_path[self.current_parse_depth:]
		self.starting_input_index[self.current_parse_depth] = None

	def peek_parserstack(self):
		try:
			retval = self.parser_stacks[-1]
		except IndexError:
			retval = None
		return retval

	def peek_parser(self):
		try:
			retval = self.peek_parserstack().peek()
		except AttributeError:
			# print("Parser stack of stacks empty!")
			retval = None

		# retval will also be None when parser stack is empty (while parser stack of stacks isn't)
		return retval

	def parser_by_address(self, parser_addr):
		try:
			return self.parser_objs[int(parser_addr)]
		except KeyError:
			print("Parser with address " + str(hex(parser_addr)) + " not found!")

	def parsers_by_name(self, parser_name):
		results = [v for k,v in self.parser_objs.items() if v.name == parser_name]
		if len(results) > 0:
			return results
		else:
			return None

	def set_input_chunk(self, chunk):
		self.input_chunk = chunk

	def get_input_chunk(self):
		return self.input_chunk

	def set_parser_env(self, parser_env):
		self.current_parser_env = parser_env

	def get_parser_env(self):
		return self.current_parser_env

	def add_or_get_parser(self, parser_addr):
		try:
			parser_obj = self.parser_objs[int(parser_addr)]
		except KeyError:
			# Create a parser object with no name and the address of the parser
			parser_obj = Parser(None, int(parser_addr))
			self.parser_objs[int(parser_addr)] = parser_obj

		return parser_obj

	def get_parser_top_per_arena_mem(self):
		return sorted(self.parser_objs.values(), key=Parser.get_arenamax, reverse=True)[0]

	def get_parser_top_total_arena_mem(self):
		return sorted(self.parser_objs.values(), key=Parser.get_arenasum, reverse=True)[0]

	def get_avg_mem_use_per_arena(self):
		avg_mem_use = {}
		arena_counts = {}
		# Accumulate byte counts and counts of parsers using that arena
		for p in self.parser_objs.values():
			for arena,mem in p.bytes_used.items(): # TODO: state vs arena
				if arena in arena_counts.keys():
					arena_counts[arena] += 1
				else:
					arena_counts[arena] = 1

				if arena in avg_mem_use.keys():
					avg_mem_use[arena] += mem
				else:
					avg_mem_use[arena] = mem

		averages = {arena: mem/arena_counts[arena] for arena,mem in avg_mem_use.items()}

		return averages

	def dump_memory_stats(self):
		return [(p.name, hex(p.address), p.bytes_used) for p in self.parser_objs.values()]

	# TODO: better naming. this is for the AST view
	# This will only work as advertised if called from h_do_parse()
	def want_result_of(self, parser=None):
		# Check that:
		# - The relevant parser stack is active (h_packrat_parse hasn't been called again, or has returned since)
		# - A particular parser application on the stack has been selected
		# - The selected parser has just been popped off the stack:
		#	- the size of the stack is equal to the "stack index" (meaning it points to one element past the top of the stack, meaning the parser of interest has just been popped)
		#	- (Potential modification in the future: the selected application is on the top of the stack and the same parser is being applied)
		if self.ast_selected_stack and self.peek_parserstack() == self.ast_selected_stack and self.ast_stack_index and len(self.ast_selected_stack.p_stack) == self.ast_stack_index:
			return True
		else:
			return False

	def setup_ast_stack_index(self, stop_index):
		self.ast_selected_stack = self.peek_parserstack()
		# Indexing from the bottom of the stack
		if stop_index < 0:
			# Convert negative index to positive one
			self.ast_stack_index = abs(stop_index)-1
		# Indexing from the top of the stack
		elif stop_index > 0:
			# The "current" parser is not on the stack while we're in h_do_parse, as it gets pushed in perform_lowlevel_parse. Consider:
			# [0x5555555779f0] digit [current] <--- not on stack, for hammer-parse-step-to-parse-result, this is #0, or #-7

			#[0x55555557f1b0] (Unnamed sequence) <--- self.ast_selected_stack.p_stack[5], #1 or #-6
			#[0x55555557f200] xrefs
			#[0x5555555800f0] (Unnamed sequence)
			#[0x555555580140] xr_td
			#[0x555555585670] (Unnamed choice)
			#[0x55555559bb90] (Unnamed ignoreseq) <--- self.ast_selected_stack.p_stack[0], #6 or #-1

			self.ast_stack_index = len(self.ast_selected_stack.p_stack)-stop_index
		# 0 has been passed in: Select current parser
		else:
			self.ast_stack_index = len(self.ast_selected_stack.p_stack)

	def clear_ast_stack_index(self):
		self.ast_stack_counter = None
		self.ast_selected_stack = None

	# TODO: get_avg_mem_use_all_arenas, get_total_mem_use
	def print_input_chunk(self, token, rows, start, end):
		print("print_input_chunk(): start: ", start, "end: ", end)
		#h = min(rows, len(token.children))
		tokenmap_values = list(self.input_token_map.values())
		#tokenmap_val_list = [token for token in list(self.input_token_map.values()) if (token['start'] <= start and token['start'] <= end)]
		tokenmap_val_list = [dict(token, addr=hex(key)) for key, token in self.input_token_map.items() if (token['start'] >= start and token['start'] <= end)]
		numrows = min(rows, len(tokenmap_val_list))
		#row_lengths_for_columns = [len([token for token in list(tokenmap_values) if (token['start'] <= x and x >= token['end'])]) for x in range(start,end+1)]
		#numrows = min(rows, max(row_lengths_for_columns))
		print("numrows: ", numrows)
		h = numrows
		w = end-start
		if w == 0:
			print("all tokens passed to display are zero-width")
			return
		charbuf = [[' '] * w for i in range(0,h)]
		print("h=len(charbuf):", len(charbuf), "w :", w)
		tokenmap_values_sorted = sorted(tokenmap_val_list, key=lambda token: (token['start'], token['end']))
		print("len(tokenmap_values_sorted):", len(tokenmap_values_sorted), "rows:", numrows)
		hpr_strings = []
		for i in range(0, numrows+1):
			token_length = tokenmap_values_sorted[i]['end'] - tokenmap_values_sorted[i]['start']
			charbuf[i][tokenmap_values_sorted[i]['start']:tokenmap_values_sorted[i]['end']] = ['X'] * min(token_length, w)
			hparseresult_addr = tokenmap_values_sorted[i]['hparseresult']
			try:
				hpr = HParseResult(int(hparseresult_addr, 16))
				hpr_strings.append(hpr.str_no_deref())
			except:
				hpr_strings.append("invalid")
#		for i in range(0, w):
#			active_tokens_unsorted = [token for token in tokenmap_values if (token['start'] <= i and i <= token['end'])]
#			# TODO: sorting this on each column seems like a lot of wasted effort
#			active_tokens = sorted(active_tokens_unsorted, key=lambda token: token['start'])
#			#print(tokenmap_values)
#			#print("len(active_tokens): ", len(active_tokens), "i: ", i)
#			#print(active_tokens)
#			for j in range(0,len(active_tokens)):
#				#print("j:", j, "j-1:", j-1)
#				#print("token_start:", active_tokens[j-1]["start"], "token_end:", active_tokens[j-1]["end"])
#				#print("i,j:", i, ",", j)
#				charbuf[j][i] = 'X'
					#try:
					#	charbuf[j-1][i] = '_'
					#except IndexError as ie:
					#	print("len(charbuf):", len(charbuf), "j: ", j)
					#	print("len(charbuf[0]):", len(charbuf[0]), "i: ", i)
					#	print("end: ", end, "start+w", start+w)
					#	print("w:", w, "h=rows:", h)
					#	raise ie
		charbufrows_token_debug = [ "".join(row) + "\n" + " ".join([hpr_strings[index], str(tokenmap_values_sorted[index])]) for index, row in enumerate(charbuf)]
		charbufrows = ["".join(row) for row in charbuf]
		#charbuf_final = "\n".join(charbufrows)
		charbuf_final = "\n".join(charbufrows_token_debug)
		print(charbuf_final)
		#print(hpr_strings)
		#print(len(hpr_strings), len(charbuf), len(tokenmap_values_sorted))

	def print_input_map(self, token, rec_depth=0, parent_bounds=None):
		w = gdb.parameter("width")
		rows = gdb.parameter("height")
		if token is None:
			print("AST pointer passed to print_input_map() is NULL")
			return
		#print(self.input_token_map)
		addr = token.address
		#addr = token
		try:
			start = self.input_token_map[int(addr)]['start']
		except KeyError as ke:
			print(self.input_token_map)
			print(ke)
		end = self.input_token_map[int(addr)]['end']
		fragmentlen = end-start
		inputfragment = (self.input_ptr+start).string('UTF-8','replace',fragmentlen)
		encoded_inputfragment = inputfragment.replace("\n", "<0A>").replace("\r", "<0D>").replace(" ", "<20>").replace("\t", "<09>")
		if rec_depth == 0:
			print(encoded_inputfragment)
			self.print_input_chunk([], rows, start, start+min(w, end-start))
		#if parent_bounds:
		#	lead = [" "] * (start-parent_bounds[0])
		#	span = ["X"] * (end-start)
		#	trail = [" "] * (parent_bounds[1]-end)
		#	final = []
		#	final.extend(lead)
		#	final.extend(span)
		#	final.extend(trail)
		#	#print(final)
		#	print("".join(final))
		#else:
		#	print("".join(['X'] * (end-start)))
		#print(start, end)
		if token.children:
			for child in token.children:
				self.print_input_map(child,rec_depth+1,(start,end))

	# TODO: better name pending
	def get_current_stack_depth(self):
		return self.current_parse_depth # TODO: what's the difference between this and ast_stack_index?

	# TODO: better name pending
	def get_top_stack_indices(self):
		return self.input_stream_indices[self.current_parse_depth:]

top_level_parse = TopLevelParse()
